from os import path
from typing import Dict
import os
from pathlib import Path
from distutils.dir_util import copy_tree
import json


def validate_attributes(title, output_directory, subtitle, report_folder_name, encryption_secret,
                        generate_encryption_secret):
    if type(title) is not str:
        raise TypeError(f'provided title is not valid. title has to be a str')
    if type(output_directory) is not str:
        raise TypeError(f'provided output_directory is not valid. output_directory has to be a str')
    if not path.exists(output_directory):
        raise NotADirectoryError(f'provided output_directory is not valid. output_directory does not exist')
    if subtitle is not None and type(subtitle) is not str:
        raise TypeError(f'provided subtitle is not valid. subtitle has to be a str')
    if report_folder_name is not None and type(report_folder_name) is not str:
        raise TypeError(f'provided report_folder_name is not valid. report_folder_name has to be a str')
    if encryption_secret is not None and type(encryption_secret) is not str:
        raise TypeError(f'provided encryption_secret is not valid. encryption_secret has to be a str')
    if encryption_secret is not None and len(encryption_secret) != 16:
        raise TypeError(f'provided encryption_secret is not valid. encryption_secret has to be 16 characters')
    if type(generate_encryption_secret) is not bool:
        raise TypeError(f'provided generate_encryption_secret is not valid. encryption_secret has to be a bool')


def _generate_encryption_secret():
    import string
    import random
    length = 16
    letters = string.ascii_lowercase + string.ascii_uppercase + string.digits + string.punctuation
    result_str = ''.join(random.choice(letters) for i in range(length))
    return result_str


def _copy_application_template(template_name: str, destination_path: str) -> None:
    package_path = str(Path(os.path.dirname(os.path.abspath(__file__))).parent)
    source_path = path.join(package_path, f'reports-templates/{template_name}')

    copy_tree(source_path, destination_path)


def _zip_directory(path: str, zip_handler):
    for root, dirs, files in os.walk(path):
        for file in files:
            zip_handler.write(os.path.join(root, file))


def _start_server_and_view_report(report_directory: str, mode: str, port: int) -> None:
    """
    Serve the report to the user using a web server.

    :param report_directory: The directory created report is saved
    :param mode: server mode ('server': will open a new tab in your default browser,
    'js': will open a new tab in your browser using a different method, 'jupyter': will open the report application
    in your notebook).
    default: 'server'
    :param port: the server port. default: random between (1024-49151)
    :return: None
    """
    print('''\n\n ### \nServing the report this way, might not work on all machine.
Try different server mode ('server', 'js' or 'jupyter') or save and download the report and open index.html \n###\n\n''')
    print('Clear your browser\'s cache if your report was not updated\n\n')
    import multiprocessing as mp
    import time
    import importlib.util
    spec = importlib.util.spec_from_file_location("app", f'{report_directory}/app.py')
    app = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(app)
    try:
        p = mp.Process(target=app.run_application, args=(port, report_directory))
        p.start()
        time.sleep(1.0)
        url = f'http://127.0.0.1:{port}/'
        if mode == 'server':
            import webbrowser
            webbrowser.open(url)
        elif mode == 'js':
            from IPython.core.display import display
            from IPython.display import Javascript
            display(Javascript('window.open("{url}");'.format(url=url)))
        else:
            from IPython.display import IFrame
            from IPython.core.display import display
            display(IFrame('http://127.0.0.1:5000/', '100%', '800px'))
        p.join()
    except (KeyboardInterrupt, SystemExit):
        print('\n! Received keyboard interrupt, stopping server.\n')
    pass


class Report:
    """
    Report is the parent class of all the reports. It contains all the method for generating the reports
    and common attributes that all reports share.

    Attributes
    ----------
    title : str
        the title of the report
    output_directory : str
        the directory where the report folder will be created
    subtitle : str default=None
        an optional subtitle to describe your report
    report_folder_name : str default=None
        the name of the folder that will contain all the generated report files.
        If not set, the title of the report will be used.
    encryption_secret : str default=None
        the secret that will be used to encrypt the generated report data.
        If it is not set, the generated data won't be encrypted.
    generate_encryption_secret : bool default=False
        the encryption_secret will be generated and its value returned as output.
        you can also view encryption_secret to get the generated secret.

    Methods
    -------
    _update_report()
        updates the report dictionary inside the report_data dictionary

    """

    def __init__(self,
                 title: str,
                 output_directory: str,
                 subtitle: str = None,
                 report_folder_name: str = None,
                 encryption_secret: str = None,
                 generate_encryption_secret: bool = False) -> None:
        super().__init__()

        validate_attributes(title,
                            output_directory,
                            subtitle,
                            report_folder_name,
                            encryption_secret,
                            generate_encryption_secret)

        self.title = title
        self.output_directory = output_directory
        self.subtitle = subtitle
        self.report_folder_name = report_folder_name
        self.encryption_secret = encryption_secret
        self.generate_encryption_secret = generate_encryption_secret

        if self.encryption_secret is None and self.generate_encryption_secret:
            self.encryption_secret = _generate_encryption_secret()

        self.report_data = {'title': title}

        if subtitle:
            self.report_data['subtitle'] = subtitle

        self.report_data['report'] = {}

    def _update_report(self, data: Dict) -> None:
        """
        Updates the report dictionary in report_data
        :param data: Dict, the data dictionary that will be added to the report dictionary in report_data
        :return: None
        """
        self.report_data['report'].update(data)

    def _serve_report_using_flask(self, template_name: str, mode: str, port: int) -> None:
        """
        Creates the report directory, copies the web application based on the template name,
        saves the report data and starts the flask server.

        :param template_name: the name of the report's template
        :param mode: the server mode
        :param port: the server port
        :return: None
        """
        report_directory = self._create_report_directory()
        _copy_application_template(template_name, report_directory)
        self._save_report_data(report_directory)
        _start_server_and_view_report(report_directory, mode, port)

    def _save_the_report(self, template_name: str, zip_report: bool) -> None:
        """
        Creates the report directory, copies the web application based on the template name,
        saves the report data.

        :param template_name: the name of the report's template
        :return: None
        """
        report_directory = self._create_report_directory()
        _copy_application_template(template_name, report_directory)
        self._save_report_data(report_directory)

        if zip_report:
            self._zip_report_directory(report_directory)

        print(f'''The report has been saved.
To view the report, go to the report's directory ({report_directory}) and open index.html then upload report_data.json.
To zip the report directory, set zip_report=True when saving.''')

    def _create_report_directory(self) -> str:
        """
        Creates the report directory if it doesn't exist.

        :return: report_directory
        """
        report_folder = self.report_folder_name if self.report_folder_name else self.title
        report_directory = path.join(self.output_directory, report_folder)
        if not os.path.exists(report_directory):
            os.mkdir(report_directory)

        return report_directory

    def _save_report_data(self, report_directory: str) -> None:
        """
        Takes the report_directory and saves the data as json there.

        :param report_directory:
        :return: None
        """
        if self.encryption_secret:
            data = self._encrypt_report_data()
            with open(path.join(report_directory, 'report_data.json'), 'wb') as file:
                file.write(data)
                file.close()
        else:
            data = self.report_data
            with open(path.join(report_directory, 'report_data.json'), 'w') as file_path:
                json.dump(data, file_path)

    def _zip_report_directory(self, report_directory: str) -> None:
        import shutil
        report_folder = self.report_folder_name if self.report_folder_name else self.title
        shutil.make_archive(f'{self.output_directory}/{report_folder}', 'zip', report_directory)

    def _encrypt_report_data(self):
        from Crypto import Random
        from Crypto.Cipher import AES
        import base64
        BLOCK_SIZE = 16

        def pad(data):
            length = 16 - (len(data) % 16)
            return data.decode("utf-8") + chr(length) * length

        def encrypt(message, passphrase):
            IV = Random.new().read(BLOCK_SIZE)
            aes = AES.new(passphrase, AES.MODE_CFB, IV, segment_size=128)
            return base64.b64encode(IV + aes.encrypt(pad(message).encode("utf-8")))

        key = self.encryption_secret.encode('utf-8')
        encoded_data = json.dumps(self.report_data).encode('utf-8')

        encrypted_data = encrypt(encoded_data, key)

        return encrypted_data



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['olliepy',
 'olliepy.utils',
 'reports-templates',
 'reports-templates.regression-error-analysis-report']

package_data = \
{'': ['*'],
 'reports-templates.regression-error-analysis-report': ['css/*',
                                                        'fonts/*',
                                                        'img/*',
                                                        'js/*']}

install_requires = \
['flask>=1.1.0',
 'ipython>=7.0.0',
 'numpy>=1.17.0',
 'pandas>=0.25.0',
 'pycrypto>=2.6.1',
 'scikit-learn>=0.22',
 'scipy>=1.3.0']

setup_kwargs = {
    'name': 'olliepy',
    'version': '0.1.11',
    'description': 'An interactive reporting tool written in python for machine learning experiments that generates interactive reports written in VueJS.',
    'long_description': None,
    'author': 'ahmed.mohamed',
    'author_email': 'hanoush87@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6',
}


setup(**setup_kwargs)



import inspect
import typing
# Credit: https://stackoverflow.com/users/1222951/aran-fey
# https://stackoverflow.com/questions/55503673/how-do-i-check-if-a-value-matches-a-type-in-python

__all__ = ['is_instance', 'is_subtype', 'python_type', 'is_generic', 'is_base_generic', 'is_qualified_generic']


if hasattr(typing, '_GenericAlias'):
    # python 3.7
    def _is_generic(cls):
        if isinstance(cls, typing._GenericAlias):
            return True

        if isinstance(cls, typing._SpecialForm):
            return cls not in {typing.Any}

        return False


    def _is_base_generic(cls):
        if isinstance(cls, typing._GenericAlias):
            if cls.__origin__ in {typing.Generic, typing._Protocol}:
                return False

            if isinstance(cls, typing._VariadicGenericAlias):
                return True

            return len(cls.__parameters__) > 0

        if isinstance(cls, typing._SpecialForm):
            return cls._name in {'ClassVar', 'Union', 'Optional'}

        return False


    def _get_base_generic(cls):
        # subclasses of Generic will have their _name set to None, but
        # their __origin__ will point to the base generic
        if cls._name is None:
            return cls.__origin__
        else:
            return getattr(typing, cls._name)


    def _get_python_type(cls):
        """
        Like `python_type`, but only works with `typing` classes.
        """
        return cls.__origin__


    def _get_name(cls):
        return cls._name
else:
    # python <3.7
    if hasattr(typing, '_Union'):
        # python 3.6
        def _is_generic(cls):
            if isinstance(cls, (typing.GenericMeta, typing._Union, typing._Optional, typing._ClassVar)):
                return True

            return False


        def _is_base_generic(cls):
            if isinstance(cls, (typing.GenericMeta, typing._Union)):
                return cls.__args__ in {None, ()}

            if isinstance(cls, typing._Optional):
                return True

            return False
    else:
        # python 3.5
        def _is_generic(cls):
            if isinstance(cls, (typing.GenericMeta, typing.UnionMeta, typing.OptionalMeta, typing.CallableMeta, typing.TupleMeta)):
                return True

            return False


        def _is_base_generic(cls):
            if isinstance(cls, typing.GenericMeta):
                return all(isinstance(arg, typing.TypeVar) for arg in cls.__parameters__)

            if isinstance(cls, typing.UnionMeta):
                return cls.__union_params__ is None

            if isinstance(cls, typing.TupleMeta):
                return cls.__tuple_params__ is None

            if isinstance(cls, typing.CallableMeta):
                return cls.__args__ is None

            if isinstance(cls, typing.OptionalMeta):
                return True

            return False


    def _get_base_generic(cls):
        try:
            return cls.__origin__
        except AttributeError:
            pass

        name = type(cls).__name__
        if not name.endswith('Meta'):
            raise NotImplementedError("Cannot determine base of {}".format(cls))

        name = name[:-4]
        return getattr(typing, name)


    def _get_python_type(cls):
        """
        Like `python_type`, but only works with `typing` classes.
        """
        # Many classes actually reference their corresponding abstract base class from the abc module
        # instead of their builtin variant (i.e. typing.List references MutableSequence instead of list).
        # We're interested in the builtin class (if any), so we'll traverse the MRO and look for it there.
        for typ in cls.mro():
            if typ.__module__ == 'builtins' and typ is not object:
                return typ

        try:
            return cls.__extra__
        except AttributeError:
            pass

        if is_qualified_generic(cls):
            cls = get_base_generic(cls)

        if cls is typing.Tuple:
            return tuple

        raise NotImplementedError("Cannot determine python type of {}".format(cls))


    def _get_name(cls):
        try:
            return cls.__name__
        except AttributeError:
            return type(cls).__name__[1:]


if hasattr(typing.List, '__args__'):
    # python 3.6+
    def _get_subtypes(cls):
        subtypes = cls.__args__

        if get_base_generic(cls) is typing.Callable:
            if len(subtypes) != 2 or subtypes[0] is not ...:
                subtypes = (subtypes[:-1], subtypes[-1])

        return subtypes
else:
    # python 3.5
    def _get_subtypes(cls):
        if isinstance(cls, typing.CallableMeta):
            if cls.__args__ is None:
                return ()

            return cls.__args__, cls.__result__

        for name in ['__parameters__', '__union_params__', '__tuple_params__']:
            try:
                subtypes = getattr(cls, name)
                break
            except AttributeError:
                pass
        else:
            raise NotImplementedError("Cannot extract subtypes from {}".format(cls))

        subtypes = [typ for typ in subtypes if not isinstance(typ, typing.TypeVar)]
        return subtypes


def is_generic(cls):
    """
    Detects any kind of generic, for example `List` or `List[int]`. This includes "special" types like
    Union and Tuple - anything that's subscriptable, basically.
    """
    return _is_generic(cls)


def is_base_generic(cls):
    """
    Detects generic base classes, for example `List` (but not `List[int]`)
    """
    return _is_base_generic(cls)


def is_qualified_generic(cls):
    """
    Detects generics with arguments, for example `List[int]` (but not `List`)
    """
    return is_generic(cls) and not is_base_generic(cls)


def get_base_generic(cls):
    if not is_qualified_generic(cls):
        raise TypeError('{} is not a qualified Generic and thus has no base'.format(cls))

    return _get_base_generic(cls)


def get_subtypes(cls):
    return _get_subtypes(cls)


def _instancecheck_iterable(iterable, type_args):
    if len(type_args) != 1:
        raise TypeError("Generic iterables must have exactly 1 type argument; found {}".format(type_args))

    type_ = type_args[0]
    return all(is_instance(val, type_) for val in iterable)


def _instancecheck_mapping(mapping, type_args):
    return _instancecheck_itemsview(mapping.items(), type_args)


def _instancecheck_itemsview(itemsview, type_args):
    if len(type_args) != 2:
        raise TypeError("Generic mappings must have exactly 2 type arguments; found {}".format(type_args))

    key_type, value_type = type_args
    return all(is_instance(key, key_type) and is_instance(val, value_type) for key, val in itemsview)


def _instancecheck_tuple(tup, type_args):
    if len(tup) != len(type_args):
        return False

    return all(is_instance(val, type_) for val, type_ in zip(tup, type_args))


_ORIGIN_TYPE_CHECKERS = {}
for class_path, check_func in {
    # iterables
    'typing.Container': _instancecheck_iterable,
    'typing.Collection': _instancecheck_iterable,
    'typing.AbstractSet': _instancecheck_iterable,
    'typing.MutableSet': _instancecheck_iterable,
    'typing.Sequence': _instancecheck_iterable,
    'typing.MutableSequence': _instancecheck_iterable,
    'typing.ByteString': _instancecheck_iterable,
    'typing.Deque': _instancecheck_iterable,
    'typing.List': _instancecheck_iterable,
    'typing.Set': _instancecheck_iterable,
    'typing.FrozenSet': _instancecheck_iterable,
    'typing.KeysView': _instancecheck_iterable,
    'typing.ValuesView': _instancecheck_iterable,
    'typing.AsyncIterable': _instancecheck_iterable,

    # mappings
    'typing.Mapping': _instancecheck_mapping,
    'typing.MutableMapping': _instancecheck_mapping,
    'typing.MappingView': _instancecheck_mapping,
    'typing.ItemsView': _instancecheck_itemsview,
    'typing.Dict': _instancecheck_mapping,
    'typing.DefaultDict': _instancecheck_mapping,
    'typing.Counter': _instancecheck_mapping,
    'typing.ChainMap': _instancecheck_mapping,

    # other
    'typing.Tuple': _instancecheck_tuple,
}.items():
    try:
        cls = eval(class_path)
    except AttributeError:
        continue

    _ORIGIN_TYPE_CHECKERS[cls] = check_func


def _instancecheck_callable(value, type_):
    if not callable(value):
        return False

    if is_base_generic(type_):
        return True

    param_types, ret_type = get_subtypes(type_)
    sig = inspect.signature(value)

    missing_annotations = []

    if param_types is not ...:
        if len(param_types) != len(sig.parameters):
            return False

        # FIXME: add support for TypeVars

        # if any of the existing annotations don't match the type, we'll return False.
        # Then, if any annotations are missing, we'll throw an exception.
        for param, expected_type in zip(sig.parameters.values(), param_types):
            param_type = param.annotation
            if param_type is inspect.Parameter.empty:
                missing_annotations.append(param)
                continue

            if not is_subtype(param_type, expected_type):
                return False

    if sig.return_annotation is inspect.Signature.empty:
        missing_annotations.append('return')
    else:
        if not is_subtype(sig.return_annotation, ret_type):
            return False

    if missing_annotations:
        raise ValueError("Missing annotations: {}".format(missing_annotations))

    return True


def _instancecheck_union(value, type_):
    types = get_subtypes(type_)
    return any(is_instance(value, typ) for typ in types)


def _instancecheck_type(value, type_):
    # if it's not a class, return False
    if not isinstance(value, type):
        return False

    if is_base_generic(type_):
        return True

    type_args = get_subtypes(type_)
    if len(type_args) != 1:
        raise TypeError("Type must have exactly 1 type argument; found {}".format(type_args))

    return is_subtype(value, type_args[0])


_SPECIAL_INSTANCE_CHECKERS = {
    'Union': _instancecheck_union,
    'Callable': _instancecheck_callable,
    'Type': _instancecheck_type,
    'Any': lambda v, t: True,
}


def is_instance(obj, type_):
    if type_.__module__ == 'typing':
        if is_qualified_generic(type_):
            base_generic = get_base_generic(type_)
        else:
            base_generic = type_
        name = _get_name(base_generic)

        try:
            validator = _SPECIAL_INSTANCE_CHECKERS[name]
        except KeyError:
            pass
        else:
            return validator(obj, type_)

    if is_base_generic(type_):
        python_type = _get_python_type(type_)
        return isinstance(obj, python_type)

    if is_qualified_generic(type_):
        python_type = _get_python_type(type_)
        if not isinstance(obj, python_type):
            return False

        base = get_base_generic(type_)
        try:
            validator = _ORIGIN_TYPE_CHECKERS[base]
        except KeyError:
            raise NotImplementedError("Cannot perform isinstance check for type {}".format(type_))

        type_args = get_subtypes(type_)
        return validator(obj, type_args)

    return isinstance(obj, type_)


def is_subtype(sub_type, super_type):
    if not is_generic(sub_type):
        python_super = python_type(super_type)
        return issubclass(sub_type, python_super)

    # at this point we know `sub_type` is a generic
    python_sub = python_type(sub_type)
    python_super = python_type(super_type)
    if not issubclass(python_sub, python_super):
        return False

    # at this point we know that `sub_type`'s base type is a subtype of `super_type`'s base type.
    # If `super_type` isn't qualified, then there's nothing more to do.
    if not is_generic(super_type) or is_base_generic(super_type):
        return True

    # at this point we know that `super_type` is a qualified generic... so if `sub_type` isn't
    # qualified, it can't be a subtype.
    if is_base_generic(sub_type):
        return False

    # at this point we know that both types are qualified generics, so we just have to
    # compare their sub-types.
    sub_args = get_subtypes(sub_type)
    super_args = get_subtypes(super_type)
    return all(is_subtype(sub_arg, super_arg) for sub_arg, super_arg in zip(sub_args, super_args))


def python_type(annotation):
    """
    Given a type annotation or a class as input, returns the corresponding python class.

    Examples:

    ::
        >>> python_type(typing.Dict)
        <class 'dict'>
        >>> python_type(typing.List[int])
        <class 'list'>
        >>> python_type(int)
        <class 'int'>
    """
    try:
        mro = annotation.mro()
    except AttributeError:
        # if it doesn't have an mro method, it must be a weird typing object
        return _get_python_type(annotation)

    if typing.Type in mro:
        return annotation.python_type
    elif annotation.__module__ == 'typing':
        return _get_python_type(annotation)
    else:
        return annotation
